@@ -17,14 +17,13 @@
  *
  */
 
-
 #ifdef HAVE_CONFIG_H
 #   include "config.h"
 #endif
 
-#include <sys/types.h>
-
+#include <signal.h>
 #include <string.h>
+#include <strings.h>
 #include <stdlib.h>
 #include <stdio.h>
 #include <unistd.h>
@@ -33,73 +32,71 @@
 #include "xnconfig_memory.h"
 
 #include "utils.h"
-#include "list.h"
+#include "list_char.h"
 #include "types.h"
 #include "log.h"
 
 #include "xnconfig.h"
 
+#define LIBRARY_API_VERSION_MAJOR	2
+#define LIBRARY_API_VERSION_MINOR	0
+
 static const char *option_names[] = 	{
 						"DefaultMode", "ExcludeApp", "AddBind", "LogLevel", "SetXkbGroup", "AddLanguage",
 						"VowelLetter", "ConsonantLetter", "NoFirstLetter",
 						"SetAutoApp", "SetManualApp", "GrabMouse", "EducationMode",
-						"Version", "LayoutRememberMode", "SaveSelectionMode"
+						"Version", "LayoutRememberMode", "SaveSelectionMode",
+						"DefaultXkbGroup"
 					};
 
 static const int total_options = sizeof(option_names) / sizeof(option_names[0]);
 
-static const char *error_messages[] =
-{
-	"Operation completed succsefully",
-	"Unrecognized option detected",
-	"Invalid value for default mode specified",
-	"Invalid value for action name specified",
-	"Invalid value for modifier name specified",
-	"Invalid value for key name specified",
-	"Invalid value for log level specified",
-	"Undefined language name in group binding detected",
-	"Can't load config file",
-	"No languages specified in config file",
-	"Can't open dictionary file",
-	"Can't open proto file",
-	"Can't open language definition file",
-	"Invalid value for mouse processing mode",
-	"Invalid value for education mode",
-	"Invalid configuration file",
-	"Invalid layout remember mode",
-	"Invalid save selection mode",
-	"Can't create file",
-};
+static const char *log_types[] = {"Error", "Warning", "Log", "Debug"};
+static const char *action_names[] = {"ChangeWord", "ChangeString", "ChangeMode", 
+									"ChangeSelected", "TranslitSelected", "ChangecaseSelected",
+									"EnableLayout1", "EnableLayout2", "EnableLayout3", "EnableLayout4"};
 
-static const unsigned int total_error_messages = sizeof(error_messages) / sizeof(error_messages[0]);
+static const char *modifier_names[] = {"Shift", "Control", "Alt", "Super"};
 
 static int glob_lang = -1;
 
 static char* get_word(char **string)
 {
 	char *str = *string;
+	char quote = ' ';
 
 	while (TRUE)
 	{
-		if (*str != ' ')
-			break;
-		str++;
+		char sym = *str;
+		if (sym == ' ')
+		{
+			str++;
+			continue;
+		}
+
+		if (sym == '"' || sym == '\'')
+		{
+			quote = sym;
+			str++;
+		}
+		break;
 	}
+
 	*string = str;
 
 	int i, len = strlen(str);
 	for (i = 0; i < len; i++)
 	{
-		if (str[i] == ' ')
+		if (str[i] == quote)
+		{
+			str[i++] = NULLSYM;
 			break;
+		}
 	}
 
-	*string += i + 1;
+	*string += i;
 
-	char *word = (char *) xnmalloc(i + 1);
-	memcpy(word, str, i);
-	word[i] = NULLSYM;
-	return word;
+	return str;
 }
 
 static int find_lang_index(struct _xneur_config *p, const char *name)
@@ -113,34 +110,32 @@ static int find_lang_index(struct _xneur_config *p, const char *name)
 	return NO_LANGUAGE;
 }
 
-static enum _error parse_line(struct _xneur_config *p, char *line)
+static int get_option_index(char *option)
+{
+	for (int i = 0; i < total_options; i++)
+	{
+		if (strcmp(option, option_names[i]) == 0)
+			return i + 1;
+	}
+	return 0;
+}
+
+static void parse_line(struct _xneur_config *p, char *line)
 {
 	if (line[0] == '#')
-		return ERROR_SUCCESS;
+		return;
 
 	char *option = get_word(&line);
 
-	int i, index = 0;
-	for (i = 0; i < total_options; i++)
-	{
-		if (strcmp(option, option_names[i]) == 0)
-		{
-			index = i + 1;
-			break;
-		}
-	}
-
+	int index = get_option_index(option);
 	if (index == 0)
 	{
-		free(option);
-		return ERROR_UNRECOGNIZED_OPTION;
+		log_message(WARNING, "Unrecognized option \"%s\" detected", option);
+		return;
 	}
 
-	free(option);
-
 	char *param = get_word(&line);
 
-	enum _error error_code = ERROR_SUCCESS;
 	switch (index)
 	{
 		case 1: // Get Default Mode (Manual/Auto)
@@ -150,7 +145,7 @@ static enum _error parse_line(struct _xneur_config *p, char *line)
 			else if (strcmp(param, "Auto") == 0)
 				p->set_current_mode(p, AUTO_MODE);
 			else
-				error_code = ERROR_INVALID_DEFAULT_MODE;
+				log_message(WARNING, "Invalid value for default mode specified");
 			break;
 		}
 		case 2: // Get Applications Names
@@ -173,53 +168,40 @@ static enum _error parse_line(struct _xneur_config *p, char *line)
 				action = ACTION_TRANSLIT_SELECTED;
 			else if (strcmp(param, "ChangecaseSelected") == 0)
 				action = ACTION_CHANGECASE_SELECTED;
+			else if (strcmp(param, "EnableLayout1") == 0)
+				action = ACTION_ENABLE_LAYOUT_0;
+			else if (strcmp(param, "EnableLayout2") == 0)
+				action = ACTION_ENABLE_LAYOUT_1;
+			else if (strcmp(param, "EnableLayout3") == 0)
+				action = ACTION_ENABLE_LAYOUT_2;
+			else if (strcmp(param, "EnableLayout4") == 0)
+				action = ACTION_ENABLE_LAYOUT_3;
 			else
 			{
-				error_code = ERROR_INVALID_ACTION_NAME;
+				log_message(WARNING, "Invalid value for action name specified");
 				break;
 			}
 
-			free(param);
-
-			param = get_word(&line);
-
-			enum _hotkey_modifier modifier;
-			if (strcmp(param, "None") == 0)
-				modifier = MODIFIER_NONE;
-			else if (strcmp(param, "Shift") == 0)
-				modifier = MODIFIER_SHIFT;
-			else if (strcmp(param, "Ctrl") == 0)
-				modifier = MODIFIER_CONTROL;
-			else if (strcmp(param, "Alt") == 0)
-				modifier = MODIFIER_ALT;
-			else
+			char *str = malloc(strlen(line) * sizeof(char));
+			char *str1;
+			strcpy(str, line);
+			const char *delim = " ";
+			str1 = strtok(str, delim);
+			while (str1 != NULL)
 			{
-				error_code = ERROR_INVALID_MODIFIER_NAME;
-				break;
+				if (strcmp(str1, modifier_names[0]) == 0)
+					p->hotkeys[action].modifier1 = strdup(str1);
+				else if (strcmp(str1, modifier_names[1]) == 0)
+					p->hotkeys[action].modifier2 = strdup(str1);
+				else if (strcmp(str1, modifier_names[2]) == 0)
+					p->hotkeys[action].modifier3 = strdup(str1);
+				else if (strcmp(str1, modifier_names[3]) == 0)
+					p->hotkeys[action].modifier4 = strdup(str1);
+				else
+					p->hotkeys[action].key = strdup(str1);
+				
+				str1 = strtok(NULL, delim);
 			}
-
-			free(param);
-
-			param = get_word(&line);
-
-			enum _hotkey_key key;
-			if (strcmp(param, "None") == 0)
-				key = KEY_NONE;
-			else if (strcmp(param, "Break") == 0)
-				key = KEY_BREAK;
-			else if (strcmp(param, "ScrollLock") == 0)
-				key = KEY_SCROLL_LOCK;
-			else if (strcmp(param, "PrtSc") == 0)
-				key = KEY_PRINT_SCREEN;
-			else
-			{
-				error_code = ERROR_INVALID_KEY_NAME;
-				break;
-			}
-
-			p->hotkeys[action].modifier = modifier;
-			p->hotkeys[action].key = key;
-
 			break;
 		}
 		case 4: // Get Debug Mode
@@ -230,8 +212,15 @@ static enum _error parse_line(struct _xneur_config *p, char *line)
 				p->log_level = LOG;
 			else if (strcmp(param, "Debug") == 0)
 				p->log_level = DEBUG;
+			else if (strcmp(param, "Warning") == 0)
+				p->log_level = WARNING;
 			else
-				error_code = ERROR_INVALID_LOG_LEVEL;
+			{
+				log_message(WARNING, "Invalid value for log level specified");
+				break;
+			}
+
+			log_set_level(p->log_level);
 			break;
 		}
 		case 5:
@@ -239,14 +228,11 @@ static enum _error parse_line(struct _xneur_config *p, char *line)
 			int lang = find_lang_index(p, param);
 			if (lang == NO_LANGUAGE)
 			{
-				error_code = ERROR_UNDEFINED_LANGUAGE_NAME;
+				log_message(WARNING, "Can't find language with name %s", param);
 				break;
 			}
 
-			free(param);
-
-			param = get_word(&line);
-			p->xkb_groups[lang] = atoi(param);
+			p->xkb_groups[lang] = atoi(get_word(&line));
 			break;
 		}
 		case 6:
@@ -262,10 +248,8 @@ static enum _error parse_line(struct _xneur_config *p, char *line)
 				p->xkb_groups	= (int *) xnrealloc(p->xkb_groups, (p->total_languages + 1) * sizeof(int));
 			}
 
-			char *dir = get_word(&line);
-
 			p->languages[p->total_languages].name		= strdup(param);
-			p->languages[p->total_languages].dir		= strdup(dir);
+			p->languages[p->total_languages].dir		= strdup(get_word(&line));
 
 			p->languages[p->total_languages].temp_dicts	= NULL;
 			p->languages[p->total_languages].dicts		= NULL;
@@ -276,26 +260,21 @@ static enum _error parse_line(struct _xneur_config *p, char *line)
 			p->xkb_groups[p->total_languages]		= p->total_languages;
 
 			p->total_languages++;
-
-			free(dir);
 			break;
 		}
 		case 7: // Get Vowel Letter
 		{
-			p->languages[glob_lang].vowel_letter = xnmalloc(strlen(param));
-			strcpy(p->languages[glob_lang].vowel_letter, param);
+			p->languages[glob_lang].vowel_letter = strdup(param);
 			break;
 		}
 		case 8: // Get Consonant Letter
 		{
-			p->languages[glob_lang].consonant_letter = xnmalloc(strlen(param));
-			strcpy(p->languages[glob_lang].consonant_letter, param);
+			p->languages[glob_lang].consonant_letter = strdup(param);
 			break;
 		}
 		case 9: // Get No First Letter
 		{
-			p->languages[glob_lang].nofirst_letter = xnmalloc(strlen(param));
-			strcpy(p->languages[glob_lang].nofirst_letter, param);
+			p->languages[glob_lang].nofirst_letter = strdup(param);
 			break;
 		}
 		case 10: // Get Auto Processing Applications
@@ -315,7 +294,7 @@ static enum _error parse_line(struct _xneur_config *p, char *line)
 			else if (strcmp(param, "No") == 0)
 				p->mouse_processing_mode = MOUSE_GRAB_DISABLE;
 			else
-				error_code = ERROR_INVALID_MOUSE_MODE;
+				log_message(WARNING, "Invalid value for mouse processing mode specified");
 			break;
 		}
 		case 13: // Get Education Mode
@@ -325,7 +304,7 @@ static enum _error parse_line(struct _xneur_config *p, char *line)
 			else if (strcmp(param, "No") == 0)
 				p->education_mode = EDUCATION_MODE_DISABLE;
 			else
-				error_code = ERROR_INVALID_EDUCATION_MODE;
+				log_message(WARNING, "Invalid value for education mode specified");
 			break;
 		}
 		case 14: // Get config version
@@ -339,62 +318,40 @@ static enum _error parse_line(struct _xneur_config *p, char *line)
 			else if (strcmp(param, "No") == 0)
 				p->layout_remember_mode = LAYOUTE_REMEMBER_DISABLE;
 			else
-				error_code = ERROR_INVALID_LAYOUTE_REMEMBER_MODE;
+				log_message(WARNING, "Invalid value for layoute remember mode specified");
 			break;
 		}
 		case 16: // Get Save Selection Mode
 		{
 			if (strcmp(param, "Yes") == 0)
-				p->save_selection_mode = SAVESELECTION_ENABLE;
+				p->save_selection_mode = SELECTION_SAVE_ENABLED;
 			else if (strcmp(param, "No") == 0)
-				p->save_selection_mode = SAVESELECTION_DISABLE;
+				p->save_selection_mode = SELECTION_SAVE_DISABLED;
 			else
-				error_code = ERROR_INVALID_SAVESELECTION_MODE;
+				log_message(WARNING, "Invalid value for selection save mode specified");
+			break;
+		}
+		case 17: // Get Initial Xkb Group for all new windows
+		{
+			p->default_group = atoi(get_word(&param));
 			break;
 		}
 	}
-
-	free(param);
-	return error_code;
 }
 
-static enum _error parse_config_file(struct _xneur_config *p, const char *dir_name, const char *file_name)
+static int parse_config_file(struct _xneur_config *p, const char *dir_name, const char *file_name)
 {
-	struct _list *list = load_list_from_file(dir_name, file_name);
+	struct _list_char *list = load_list_from_file(dir_name, file_name);
 	if (list == NULL)
-		return ERROR_CANT_LOAD_FILE;
-
-	enum _error error_code = ERROR_SUCCESS;
-	int i;
-	for (i = 0; i < list->data_count; i++)
 	{
-		error_code = parse_line(p, list->data[i].string);
-		if (error_code != ERROR_SUCCESS)
-			break;
+		log_message(ERROR, "Can't find config file %s", file_name);
+		return FALSE;
 	}
 
+	for (int i = 0; i < list->data_count; i++)
+		parse_line(p, list->data[i].string);
+
 	list->uninit(list);
-	return error_code;
-}
-
-static int is_process_exists(int pid)
-{
-	if (pid <= 0)
-		return FALSE;
-
-	int cmdline_file_name_len = 64;
-	char *cmdline_file_name = (char *) xnmalloc((cmdline_file_name_len + 1) * sizeof(char));
-
-	snprintf(cmdline_file_name, cmdline_file_name_len, "/proc/%d/cmdline", pid);
-
-	FILE *cmdline_file = fopen(cmdline_file_name, "r");
-	free(cmdline_file_name);
-
-	if (!cmdline_file)
-		return FALSE;
-
-	fclose(cmdline_file);
-
 	return TRUE;
 }
 
@@ -410,19 +367,44 @@ static int check_memory_attached(struct _xneur_config *p)
 	return TRUE;
 }
 
+void xneur_reload_config(struct _xneur_config *p)
+{
+	int xneur_pid = p->xneur_data->xneur_pid;
+	if (xneur_pid <= 0)
+		return;
+
+	kill(xneur_pid, SIGHUP);
+}
+
 void xneur_config_set_pid(struct _xneur_config *p, int pid)
 {
 	p->xneur_data->xneur_pid = pid;
 }
 
+int xneur_kill(struct _xneur_config *p)
+{
+	int xneur_pid = p->xneur_data->xneur_pid;
+	if (xneur_pid <= 0)
+		return FALSE;
+
+	if (kill(xneur_pid, SIGTERM) == -1)
+		return FALSE;
+
+	xneur_config_set_pid(p, 0);
+
+	return TRUE;
+}
+
 int xneur_config_get_pid(struct _xneur_config *p)
 {
 	int xneur_pid = p->xneur_data->xneur_pid;
-
-	if (is_process_exists(xneur_pid) != TRUE)
+	if (xneur_pid <= 0)
 		return -1;
 
-	return xneur_pid;
+	if (getsid(xneur_pid) == -1)
+		return -1;
+
+	return p->xneur_data->xneur_pid;
 }
 
 void xneur_set_current_mode(struct _xneur_config *p, int mode)
@@ -435,261 +417,188 @@ int xneur_get_current_mode(struct _xneur_config *p)
 	return p->xneur_data->xneur_mode;
 }
 
-enum _error xneur_load(struct _xneur_config *p)
+int xneur_load(struct _xneur_config *p)
 {
-	enum _error error_code = parse_config_file(p, NULL, CONFIG_NAME);
-	if (error_code != ERROR_SUCCESS)
-		return error_code;
+	if (!parse_config_file(p, NULL, CONFIG_NAME))
+		return FALSE;
 
 	if (p->total_languages == 0)
-		return ERROR_NO_LANGUAGES;
+	{
+		log_message(ERROR, "No languages specified in config file");
+		return FALSE;
+	}
 
-	int lang;
-	for (lang = 0; lang < p->total_languages; lang++)
+	for (int lang = 0; lang < p->total_languages; lang++)
 	{
 		char *lang_dir = p->get_lang_dir(p, lang);
-
-		p->languages[lang].temp_dicts = load_list_from_file(lang_dir, DICT_NAME);
-		if (p->languages[lang].temp_dicts == NULL)
-			return ERROR_DICT_NOT_FOUND;
+		char *lang_name = p->get_lang_name(p, lang);
 
 		p->languages[lang].dicts = load_list_from_file(lang_dir, DICT_NAME);
 		if (p->languages[lang].dicts == NULL)
-			return ERROR_DICT_NOT_FOUND;
+		{
+			log_message(ERROR, "Can't find dictionary file for %s language", lang_name);
+			return FALSE;
+		}
 
 		p->languages[lang].protos = load_list_from_file(lang_dir, PROTO_NAME);
 		if (p->languages[lang].protos == NULL)
-			return ERROR_PROTO_NOT_FOUND;
+		{
+			log_message(ERROR, "Can't find protos file for %s language", lang_name);
+			return FALSE;
+		}
 
 		p->languages[lang].big_protos = load_list_from_file(lang_dir, BIG_PROTO_NAME);
 		if (p->languages[lang].big_protos == NULL)
-			return ERROR_PROTO_NOT_FOUND;
+		{
+			log_message(ERROR, "Can't find big protos file for %s language", lang_name);
+			return FALSE;
+		}
 
 		p->languages[lang].regexp = load_list_from_file(lang_dir, REGEXP_NAME);
 		if (p->languages[lang].regexp == NULL)
-			return ERROR_PROTO_NOT_FOUND;
+		{
+			log_message(ERROR, "Can't find regexp file for %s language", lang_name);
+			return FALSE;
+		}
+
+		p->languages[lang].temp_dicts = p->languages[lang].dicts->clone(p->languages[lang].dicts);
 
 		glob_lang = lang;
-		error_code = parse_config_file(p, lang_dir, LANGDEF_NAME);
-		if (error_code != ERROR_SUCCESS)
-			return error_code;
+		if (!parse_config_file(p, lang_dir, LANGDEF_NAME))
+			return FALSE;
 	}
 
-	return ERROR_SUCCESS;
+	return TRUE;
 }
 
-enum _error xneur_save(struct _xneur_config *p)
+int xneur_save(struct _xneur_config *p)
 {
-	char *config_file_path_name = get_file_path_name(NULL, CONFIG_NAME);
+ 	char *config_file_path_name = get_file_path_name(NULL, CONFIG_NAME);
 
 	log_message(LOG, "Saving main config to %s", config_file_path_name);
 
 	FILE *stream = fopen(config_file_path_name, "w");
+	if (stream == NULL)
+	{
+		log_message(ERROR, "Can't create file %s", config_file_path_name);
+		free(config_file_path_name);
+		return FALSE;
+	}
 
 	free(config_file_path_name);
 
-	if (stream == NULL)
-		return ERROR_CANT_CREATE_FILE;
-
-	fputs("# It's a X Neural Switcher configuration file by XNeur\n# All values writted XNeur\n\n", stream);
-
-	fprintf(stream, "# Config version \nVersion %s\n\n", VERSION);
+	fprintf(stream, "# It's a X Neural Switcher configuration file by XNeur\n# All values writted XNeur\n\n");
+	fprintf(stream, "# Config version\nVersion %s\n\n", VERSION);
 
 	if (p->xneur_data->xneur_mode == AUTO_MODE)
 		fprintf(stream, "# Default work mode\nDefaultMode Auto\n\n");
 	else
 		fprintf(stream, "# Default work mode\nDefaultMode Manual\n\n");
 
-	fputs("# Level of messages program will write to output\n", stream);
-	fputs("# Example:\n", stream);
-	fputs("#LogLevel Error\n", stream);
-	fputs("#LogLevel Log\n", stream);
-	fputs("#LogLevel Debug\n", stream);
-	switch (p->log_level)
-	{
-		case LOG:
-		{
-			fprintf(stream, "LogLevel Log\n\n");
-			break;
-		}
-		case DEBUG:
-		{
-			fprintf(stream, "LogLevel Debug\n\n");
-			break;
-		}
-		case ERROR:
-		{
-			fprintf(stream, "LogLevel Error\n\n");
-			break;
-		}
-	}
+	fprintf(stream, "# Level of messages program will write to output\n");
+	fprintf(stream, "#LogLevel Error\n");
+	fprintf(stream, "#LogLevel Warning\n");
+	fprintf(stream, "#LogLevel Log\n");
+	fprintf(stream, "#LogLevel Debug\n");
+	fprintf(stream, "LogLevel %s\n", log_types[p->log_level]);
+	fprintf(stream, "\n");
+	fprintf(stream, "# Define used languages\n");
+	fprintf(stream, "# See Settings page on http://www.xneur.ru for details\n");
 
-	fputs("# Define used languages\n", stream);
-	fputs("# See Settings page on http://www.xneur.ru for details\n", stream);
-
-	int lang;
-	for (lang = 0; lang < p->total_languages; lang++)
+	for (int lang = 0; lang < p->total_languages; lang++)
 	{
 		fprintf(stream, "AddLanguage %s %s\n", p->languages[lang].name, p->languages[lang].dir);
 		fprintf(stream, "SetXkbGroup %s %d\n", p->languages[lang].name, p->xkb_groups[lang]);
 	}
-	fputs("\n", stream);
+	fprintf(stream, "\n");
 
-	fputs("# Add Applications names to exclude it from procces with xneur\n", stream);
-	fputs("# Xneur will not process the input for this applications\n", stream);
-	fputs("# Example:\n", stream);
-	fputs("#ExcludeApp Gaim\n", stream);
+	fprintf(stream, "# Define initial keyboard layout for all new applications\n");
+	fprintf(stream, "DefaultXkbGroup %d\n", p->default_group);
+	fprintf(stream, "\n");
+	
+	fprintf(stream, "# Add Applications names to exclude it from procces with xneur\n");
+	fprintf(stream, "# Xneur will not process the input for this applications\n");
+	fprintf(stream, "# Example:\n");
+	fprintf(stream, "#ExcludeApp Gaim\n");
 	for (int i = 0; i < p->excluded_apps->data_count; i++)
 		fprintf(stream, "ExcludeApp %s\n", p->excluded_apps->data[i].string);
-	fputs("\n", stream);
+	fprintf(stream, "\n");
 
-	fputs("# Use this parameter to force set work mode in current application to Auto.\n", stream);
-	fputs("# Example:\n", stream);
-	fputs("#SetAutoApp Gedit\n", stream);
+	fprintf(stream, "# Use this parameter to force set work mode in current application to Auto.\n");
+	fprintf(stream, "# Example:\n");
+	fprintf(stream, "#SetAutoApp Gedit\n");
 	for (int i = 0; i < p->auto_apps->data_count; i++)
 		fprintf(stream, "SetAutoApp %s\n", p->auto_apps->data[i].string);
-	fputs("\n", stream);
+	fprintf(stream, "\n");
 
-	fputs("# Use this parameter to force set work mode in current application to Manual.\n", stream);
-	fputs("# Example:\n", stream);
-	fputs("#SetManualApp Anjuta\n", stream);
+	fprintf(stream, "# Use this parameter to force set work mode in current application to Manual.\n");
+	fprintf(stream, "# Example:\n");
+	fprintf(stream, "#SetManualApp Anjuta\n");
 	for (int i = 0; i < p->manual_apps->data_count; i++)
 		fprintf(stream, "SetManualApp %s\n", p->manual_apps->data[i].string);
-	fputs("\n", stream);
+	fprintf(stream, "\n");
 
-	fputs("# Binds hotkeys for some actions\n", stream);
+	fprintf(stream, "# Binds hotkeys for some actions\n");
 	for (int action = 0; action < MAX_HOTKEYS; action++)
 	{
-		switch (action)
-		{
-			case ACTION_CHANGE_WORD:
-			{
-				fputs("AddBind ChangeWord ", stream);
-				break;
-			}
-			case ACTION_CHANGE_STRING:
-			{
-				fputs("AddBind ChangeString ", stream);
-				break;
-			}
-			case ACTION_CHANGE_MODE:
-			{
-				fputs("AddBind ChangeMode ", stream);
-				break;
-			}
-			case ACTION_CHANGE_SELECTED:
-			{
-				fputs("AddBind ChangeSelected ", stream);
-				break;
-			}
-			case ACTION_TRANSLIT_SELECTED:
-			{
-				fputs("AddBind TranslitSelected ", stream);
-				break;
-			}
-			case ACTION_CHANGECASE_SELECTED:
-			{
-				fputs("AddBind ChangecaseSelected ", stream);
-				break;
-			}
-		}
-
-		switch (p->hotkeys[action].modifier)
-		{
-			case MODIFIER_NONE:
-			{
-				fputs("None ", stream);
-				break;
-			}
-			case MODIFIER_SHIFT:
-			{
-				fputs("Shift ", stream);
-				break;
-			}
-			case MODIFIER_CONTROL:
-			{
-				fputs("Ctrl ", stream);
-				break;
-			}
-			case MODIFIER_ALT:
-			{
-				fputs("Alt ", stream);
-				break;
-			}
-		}
-
-		switch (p->hotkeys[action].key)
-		{
-			case KEY_NONE:
-			{
-				fputs("None \n", stream);
-				break;
-			}
-			case KEY_BREAK:
-			{
-				fputs("Break \n", stream);
-				break;
-			}
-			case KEY_SCROLL_LOCK:
-			{
-				fputs("ScrollLock \n", stream);
-				break;
-			}
-			case KEY_PRINT_SCREEN:
-			{
-				fputs("PrtSc \n", stream);
-				break;
-			}
-		}
+		fprintf(stream, "AddBind %s ", action_names[action]);
+		if (p->hotkeys[action].modifier1 != NULL)
+			fprintf(stream, "%s ", p->hotkeys[action].modifier1);
+		if (p->hotkeys[action].modifier2 != NULL)
+			fprintf(stream, "%s ", p->hotkeys[action].modifier2);
+		if (p->hotkeys[action].modifier3 != NULL)
+			fprintf(stream, "%s ", p->hotkeys[action].modifier3);
+		if (p->hotkeys[action].modifier4 != NULL)
+			fprintf(stream, "%s ", p->hotkeys[action].modifier4);
+		fprintf(stream, "%s\n", p->hotkeys[action].key);
 	}
-	fputs("\n", stream);
+	fprintf(stream, "\n");
 
-	fputs("# This option enable or disable mouse processing\n", stream);
-	fputs("# Example:\n", stream);
-	fputs("#GrabMouse Yes\n", stream);
+	fprintf(stream, "# This option enable or disable mouse processing\n");
+	fprintf(stream, "# Example:\n");
+	fprintf(stream, "#GrabMouse Yes\n");
 	if (p->mouse_processing_mode)
 		fprintf(stream, "GrabMouse Yes\n\n");
 	else
 		fprintf(stream, "GrabMouse No\n\n");
 
-	fputs("# This option enable or disable self education of xneur\n", stream);
-	fputs("# Example:\n", stream);
-	fputs("#EducationMode No\n", stream);
+	fprintf(stream, "# This option enable or disable self education of xneur\n");
+	fprintf(stream, "# Example:\n");
+	fprintf(stream, "#EducationMode No\n");
 	if (p->education_mode)
 		fprintf(stream, "EducationMode Yes\n\n");
 	else
 		fprintf(stream, "EducationMode No\n\n");
 
-	fputs("# This option enable or disable layout remember for each window\n", stream);
-	fputs("# Example:\n", stream);
-	fputs("#LayoutRememberMode No\n", stream);
+	fprintf(stream, "# This option enable or disable layout remember for each window\n");
+	fprintf(stream, "# Example:\n");
+	fprintf(stream, "#LayoutRememberMode No\n");
 	if (p->layout_remember_mode)
 		fprintf(stream, "LayoutRememberMode Yes\n\n");
 	else
 		fprintf(stream, "LayoutRememberMode No\n\n");
 
-	fputs("# This option enable or disable saving selection text\n", stream);
-	fputs("# Example:\n", stream);
-	fputs("#SaveSelectionMode No\n", stream);
+	fprintf(stream, "# This option enable or disable saving selection text\n");
+	fprintf(stream, "# Example:\n");
+	fprintf(stream, "#SaveSelectionMode No\n");
 	if (p->save_selection_mode)
 		fprintf(stream, "SaveSelectionMode Yes\n\n");
 	else
 		fprintf(stream, "SaveSelectionMode No\n\n");
 
-	fputs("# That's all\n", stream);
+	fprintf(stream, "# That's all\n");
 
 	fclose(stream);
 
-	return ERROR_SUCCESS;
+	return TRUE;
 }
 
-enum _error xneur_save_dicts(struct _xneur_config *p)
+void xneur_save_dicts(struct _xneur_config *p)
 {
 	if (p->education_mode == EDUCATION_MODE_DISABLE)
-		return ERROR_SUCCESS;
+		return;
 
-	int lang;
-	for (lang = 0; lang < p->total_languages; lang++)
+	for (int lang = 0; lang < p->total_languages; lang++)
 	{
 		char *dict_path = get_file_path_name(p->get_lang_dir(p, lang), DICT_NAME);
 
@@ -699,15 +608,6 @@ enum _error xneur_save_dicts(struct _xneur_config *p)
 
 		free(dict_path);
 	}
-
-	return ERROR_SUCCESS;
-}
-
-const char* format_error_message(enum _error error_code)
-{
-	if (error_code >= total_error_messages)
-		return "Unrecognized error";
-	return error_messages[error_code];
 }
 
 char* xneur_config_get_lang_dir(struct _xneur_config *p, int lang)
@@ -726,8 +626,7 @@ char* xneur_config_get_lang_name(struct _xneur_config *p, int lang)
 
 int xneur_config_find_group_lang(struct _xneur_config *p, int group)
 {
-	int lang;
-	for (lang = 0; lang < p->total_languages; lang++)
+	for (int lang = 0; lang < p->total_languages; lang++)
 	{
 		if (p->xkb_groups[lang] == group)
 			return lang;
@@ -744,8 +643,7 @@ int xneur_config_get_lang_group(struct _xneur_config *p, int lang)
 
 void xneur_set_languages(struct _xneur_config *p, char **lang_name, char **lang_dir, int *lang_group, int total_lang)
 {
-	int lang;
-	for (lang = 0; lang < p->total_languages; lang++)
+	for (int lang = 0; lang < p->total_languages; lang++)
 	{
 		if (p->languages[lang].temp_dicts != NULL)
 			p->languages[lang].temp_dicts->uninit(p->languages[lang].temp_dicts);
@@ -775,28 +673,37 @@ void xneur_set_languages(struct _xneur_config *p, char **lang_name, char **lang_
 	p->total_languages = total_lang;
 	p->languages	= (struct _xneur_language *) xnmalloc(p->total_languages * sizeof(struct _xneur_language));
 	p->xkb_groups	= (int *) xnmalloc(p->total_languages * sizeof(int));
-	for (lang = 0; lang < p->total_languages; lang++)
+
+	for (int lang = 0; lang < p->total_languages; lang++)
 	{
-		p->languages[lang].name	= strdup(lang_name[lang]);
-		p->languages[lang].dir	= strdup(lang_dir[lang]);
+		p->languages[lang].name		= strdup(lang_name[lang]);
+		p->languages[lang].dir		= strdup(lang_dir[lang]);
 
-		p->languages[lang].temp_dicts		= NULL;
-		p->languages[lang].dicts			= NULL;
-		p->languages[lang].protos			= NULL;
-		p->languages[lang].big_protos		= NULL;
-		p->languages[lang].regexp			= NULL;
+		p->languages[lang].temp_dicts	= NULL;
+		p->languages[lang].dicts	= NULL;
+		p->languages[lang].protos	= NULL;
+		p->languages[lang].big_protos	= NULL;
+		p->languages[lang].regexp	= NULL;
 
-		p->xkb_groups[lang]	= lang_group[lang];
+		p->xkb_groups[lang]		= lang_group[lang];
 	}
 
 }
 
+void xneur_get_library_api_version(int *major_version, int *minor_version)
+{
+	*major_version = LIBRARY_API_VERSION_MAJOR;
+	*minor_version = LIBRARY_API_VERSION_MINOR;
+}
+
 void xneur_config_uninit(struct _xneur_config *p)
 {
+	p->window_layouts->uninit(p->window_layouts);
+	p->manual_apps->uninit(p->manual_apps);
+	p->auto_apps->uninit(p->auto_apps);
 	p->excluded_apps->uninit(p->excluded_apps);
 
-	int lang;
-	for (lang = 0; lang < p->total_languages; lang++)
+	for (int lang = 0; lang < p->total_languages; lang++)
 	{
 		if (p->languages[lang].temp_dicts != NULL)
 			p->languages[lang].temp_dicts->uninit(p->languages[lang].temp_dicts);
@@ -823,6 +730,20 @@ void xneur_config_uninit(struct _xneur_config *p)
 		free(p->languages);
 	}
 
+	for (int action = 0; action < MAX_HOTKEYS; action++)
+	{
+		if (p->hotkeys[action].modifier1 != NULL)
+			free(p->hotkeys[action].modifier1);
+		if (p->hotkeys[action].modifier2 != NULL)
+			free(p->hotkeys[action].modifier2);
+		if (p->hotkeys[action].modifier3 != NULL)
+			free(p->hotkeys[action].modifier3);
+		if (p->hotkeys[action].modifier4 != NULL)
+			free(p->hotkeys[action].modifier4);
+		if (p->hotkeys[action].key != NULL)
+			free(p->hotkeys[action].key);
+	}
+		
 	free(p);
 }
 
@@ -830,19 +751,18 @@ struct _xneur_config* xneur_config_init(void)
 {
 	struct _xneur_config *p = (struct _xneur_config *) xnmalloc(sizeof(struct _xneur_config));
 	bzero(p, sizeof(struct _xneur_config));
-
-	if (check_memory_attached(p) != TRUE)
+		
+	if (!check_memory_attached(p))
 	{
 		free(p);
 		return NULL;
 	}
 
 	p->log_level			= LOG;
-	p->excluded_apps		= list_init();
-	p->auto_apps			= list_init();
-	p->manual_apps			= list_init();
-
-	p->window_layouts		= list_init();
+	p->excluded_apps		= list_char_init();
+	p->auto_apps			= list_char_init();
+	p->manual_apps			= list_char_init();
+	p->window_layouts		= list_char_init();
 
 	// Function mapping
 	p->load				= xneur_load;
@@ -856,11 +776,15 @@ struct _xneur_config* xneur_config_init(void)
 	p->get_lang_name		= xneur_config_get_lang_name;
 	p->find_group_lang		= xneur_config_find_group_lang;
 	p->get_lang_group		= xneur_config_get_lang_group;
+	p->reload_config		= xneur_reload_config;
+	p->kill				= xneur_kill;
 	p->set_languages		= xneur_set_languages;
-	p->get_file_path_name		= get_file_path_name;
+	p->get_library_api_version	= xneur_get_library_api_version;
+	p->get_dict_path		= get_file_path_name;
 	p->uninit			= xneur_config_uninit;
 
 	p->set_current_mode(p, AUTO_MODE);
-
+	p->default_group = 0;
+	
 	return p;
 }
