@@ -20,10 +20,10 @@
 #ifndef _CONFIG_FILES_H_
 #define _CONFIG_FILES_H_
 
-struct _list* load_list_from_file(const char *dir_name, const char *file_name);
+struct _list_char* load_list_from_file(const char *dir_name, const char *file_name);
 char* get_file_path_name(const char *dir_name, const char *file_name);
 char* get_file_content_path(const char *dir_name, const char *file_name);
 char* get_file_content(const char *file_name);
-void save_list_to_file(struct _list *list, const char *file_name);
+void save_list_to_file(struct _list_char *list, const char *file_name);
 
 #endif /* _CONFIG_FILES_H_ */
