@@ -22,7 +22,7 @@
 
 #define RED_COLOR		"\033[0;31m"
 #define GREEN_COLOR		"\033[0;32m"
-#define YELLOW_COLOR		"\033[0;33m"
+#define ORANGE_COLOR		"\033[0;33m"
 #define BLUE_COLOR		"\033[0;34m"
 #define TEAL_COLOR		"\033[0;36m"
 #define LIGHT_YELLOW_COLOR	"\033[1;33m"
