@@ -27,7 +27,7 @@
 
 #include "log.h"
 
-static int LOG_LEVEL = 10;
+static int LOG_LEVEL = DEBUG;
 
 void log_set_level(int level)
 {
@@ -49,10 +49,9 @@ void log_message(int level, const char *string, ...)
 			stream = stderr;
 			break;
 		}
-		default:
-		case DEBUG:
+		case WARNING:
 		{
-			modifier = TEAL_COLOR "[DBG]: " NORMAL_COLOR;
+			modifier = ORANGE_COLOR "[WRN]: " NORMAL_COLOR;
 			break;
 		}
 		case LOG:
@@ -60,6 +59,12 @@ void log_message(int level, const char *string, ...)
 			modifier = GREEN_COLOR "[LOG]: " NORMAL_COLOR;
 			break;
 		}
+		default:
+		case DEBUG:
+		{
+			modifier = TEAL_COLOR "[DBG]: " NORMAL_COLOR;
+			break;
+		}
 	}
 
 	int len = strlen(string) + strlen(modifier) + 2;
