@@ -23,4 +23,4 @@
 void log_set_level(int level);
 void log_message(int type, const char *string, ...);
 
-#endif /* _LOG_H */
+#endif /* _LOG_H_ */
