@@ -24,13 +24,13 @@
 
 void bind_manual_actions(void);
 
-enum _hotkey_action get_manual_action(KeySym key_sym, KeySym key_modifier);
+enum _hotkey_action get_manual_action(KeySym key_sym, int mask);
 
 struct _xbtable
 {
 	KeySym key_sym;
-	KeySym key_sym_alt;
-	KeySym key_modifier;
+	KeySym key_sym_shift;
+	int modifier_mask;
 };
 
 #endif /* _XBTABLE_H_ */
