@@ -22,18 +22,18 @@
 
 #include <X11/Xutil.h>
 
-#include "xkeymap.h"
-
 struct _xwindow
 {
+	struct _xkeymap *xkeymap;
+
 	Display *display;
 	Window window;
-	
-	struct _xkeymap *xkeymap;
-	
+
 	int  (*create) (struct _xwindow *p);
 	void (*destroy) (struct _xwindow *p);
 	void (*move_window) (struct _xwindow *p, int x, int y);
+	int  (*init_keymap) (struct _xwindow *p);
+	void (*uninit) (struct _xwindow *p);
 };
 
 struct _xwindow* xwindow_init(void);
